from __future__ import annotations

from traceback import format_exception_only
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandParser

if TYPE_CHECKING:
    from collections.abc import Iterable

    from django.db import models


class Command(BaseCommand):
    help = "Create admin user"

    def get_model(self) -> models.Model:
        return get_user_model()

    def get_fields(self) -> Iterable[str]:
        model = self.get_model()
        # result = []
        # for field_name in (model.USERNAME_FIELD, getattr(model.PASSWORD_FIELD, "password"), *model.REQUIRED_FIELDS):
        #     field = model._meta.get_field(field_name)
        #     if not field.null and not field.default:
        #         result.append(field_name)
        # return result
        return {model.USERNAME_FIELD, model.PASSWORD_FIELD.password, *model.REQUIRED_FIELDS}

    def add_arguments(self, parser: CommandParser) -> None:
        defaults = {"email": "admin@localhost", "username": "admin", "password": "admin"}

        for field in self.get_fields():
            kwargs = {
                "type": str,
                "required": not settings.DEBUG,
                "default": defaults.get(field, "admin") if settings.DEBUG else None,
            }
            parser.add_argument(f"-{field[0]}", f"--{field}", **kwargs)

    def handle(self, *args, **options) -> None:
        params = {k: v for k, v in options.items() if k in self.get_fields()}

        try:
            user = get_user_model().objects.create_superuser(**params)
            self.stdout.write(
                self.style.SUCCESS("Successfully created user [")
                + self.style.WARNING(f"pk={user.pk}")
                + self.style.SUCCESS("]")
            )
        except Exception as e:  # noqa: BLE001
            self.stdout.write(self.style.WARNING("".join(format_exception_only(e))))



from __future__ import annotations
from django.db import models
import json
import contextlib
import pydantic
from pydantic import BaseModel
from django.core.exceptions import ValidationError
from typing import Callable

from .serializers.json import JSONEncoder


class PydanticField(models.JSONField):
    def __init__(  # noqa: PLR0913
        self,
        model: type[BaseModel],
        verbose_name: str | None = None,
        name: str | None = None,
        encoder: type[json.JSONEncoder] | None = None,
        decoder: type[json.JSONDecoder] | None = None,
        default: Callable | None = models.NOT_PROVIDED,
        **kwargs,
    ) -> None:
        if encoder is None:
            encoder = JSONEncoder
        if default is models.NOT_PROVIDED:
            with contextlib.suppress(pydantic.ValidationError):
                model()  # модель может быть без дефолтных значений
                default = model

        super().__init__(verbose_name, name, encoder, decoder, default=default, **kwargs)
        self._pydantic_model = model

    def from_db_value(self, value, expression, connection):  # noqa: ANN001, ANN201
        v = super().from_db_value(value, expression, connection)
        if self.null and v is None:
            return None
        return self._pydantic_model.model_validate(v)

    def get_db_prep_value(self, value, connection, prepared=False):  # noqa: ANN001, ANN201, PLR0911
        if self.null and value is None:
            return None

        # нормальная pydantic модель
        if hasattr(value, "model_dump"):
            return super().get_db_prep_value(value.model_dump(), connection, prepared)

        # обёртка в виде models.Value
        if isinstance(value, models.Value):
            return self.get_db_prep_value(value.value, connection, prepared)

        # Cast и иже с ним
        if hasattr(value, "get_source_expressions"):
            if expressions := value.get_source_expressions():
                expr = expressions[0]

                # При null=True django говорит "смотри как могу" и оборачивает
                # всё в Case-When-Else, а внутри всё в Value. Прикольно, чё.
                if isinstance(expr, models.Case):
                    for when in expr.cases:
                        if isinstance(when, models.When):
                            return self.get_db_prep_value(when.result, connection, prepared)
                    if hasattr(expr, "default"):
                        return self.get_db_prep_value(expr.default, connection, prepared)

                return self.get_db_prep_value(expr.value, connection, prepared)
            return None

        return super().get_db_prep_value(value, connection, prepared)

    def to_python(self, value):  # noqa: ANN001, ANN201
        if self.null and value is None:
            return None
        try:
            return self._pydantic_model.model_validate(super().to_python(value))
        except pydantic.ValidationError as e:
            raise ValidationError(str(e)) from e

    def deconstruct(self):  # noqa: ANN201
        name, path, args, kwargs = super().deconstruct()
        return name, path, (self._pydantic_model, *args), kwargs

    def value_to_string(self, obj):  # noqa: ANN001, ANN201
        if self.null and obj is None:
            return None

        # easyaudit прокидывает сюда значение, из которого достанется словарь
        value = self.value_from_object(obj)
        if isinstance(value, dict):
            return value
        return value.model_dump()



import argparse
import os
import re
import subprocess
from traceback import format_exception_only
from typing import Any

from django.core import management
from django.core.management.base import BaseCommand


class AppendWithRelatedActions(argparse.Action):
    """
    Экшон для argparse, который позволяет привязать к аргументу подаргументы

    Он складывает неймспейсы со значениями связанных аргументов в список
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, default=[], **kwargs)
        self.related = []  # список связанных аргументов

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        dest = getattr(namespace, self.dest)
        dest.append(
            argparse.Namespace(
                type=option_string.lstrip(parser.prefix_chars).replace("-", "_"),  # не особо общий код
                value=values,
                **{action.dest: action.default for action in self.related if action.default is not None},
            )
        )


class LinkedStoreTrueAction(argparse._StoreTrueAction):
    """
    Наследоваться от приватных классов плохо, но они сами виноваты
    """

    def __init__(self, *args, linked_to: AppendWithRelatedActions, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.linked_to = linked_to
        linked_to.related.append(self)

    def get_namespace(self, namespace: argparse.Namespace) -> argparse.Namespace:
        """
        Нужен неймспейс последнего тега, с которым связан
        """
        namespaces = getattr(namespace, self.linked_to.dest)
        if not namespaces:
            raise argparse.ArgumentError(self, f"can't be used before {' or '.join(self.linked_to.option_strings)}")
        return namespaces[-1]

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        super().__call__(parser, self.get_namespace(namespace), values, option_string)


def expand_environment_variables(arg: str) -> str:
    return re.sub(r"\$(\w+)", lambda m: os.getenv(m.group(1), ""), arg)


class Command(BaseCommand):
    """
    Runs command chain.

    Команды должны быть целиком в строке, чтобы парсер не пытался обработать их агрументы:
    manage.py chain --shell 'ls -l' --manage 'collectstatic --noinput' и т.п.
    """

    help = "Chain commands execution"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        command = parser.add_argument("--manage", "--shell", action=AppendWithRelatedActions, dest="commands")
        parser.add_argument("--allow-failure", action=LinkedStoreTrueAction, linked_to=command)

    def handle(self, *args, **options) -> None:
        for command in options.get("commands", []):
            try:
                self.stdout.write(self.style.SUCCESS(f"> {command.value}"))
                if command.type == "shell":
                    subprocess.run(command.value, shell=True, check=True)
                else:
                    # а тут мы будем подставлять значения переменных окружения сами
                    argv = command.value.split()
                    expanded = (expand_environment_variables(arg) for arg in argv)
                    management.call_command(*expanded)
            except Exception as e:
                if not command.allow_failure:
                    raise

                self.stdout.write(self.style.WARNING("".join(format_exception_only(e))))



import dj_database_url
from pydantic import BaseModel, model_serializer


class Database(BaseModel):
    """
    Параметры подключения к базе данных. Основано на параметрах функции dj_database_url.parse
    """

    url: str
    engine: str | None = None
    conn_max_age: int | None = 0
    conn_health_checks: bool = False
    disable_server_side_cursors: bool = False
    ssl_require: bool = False
    test_options: dict | None = None

    @model_serializer
    def serialize(self) -> dj_database_url.DBConfig:
        return dj_database_url.parse(**{field: value for field, value in self})



import inspect
from pydantic_settings import BaseSettings, SettingsConfigDict

from .utils import alias_generator


class DjangoSettings(BaseSettings):
    """
    Базовый класс для настроек django-проекта
    """

    model_config = SettingsConfigDict(alias_generator=alias_generator)

    def __init__(self, **values):
        super().__init__(**values)
        self._set_globals()

    def _set_globals(self, offset=2) -> None:
        """
        Заполняет содержимым модели глобальные переменные модуля, в котором вызван

        offset по умолчанию 2, потому что при предполагаемом использовании метода стэк выглядит так:
            0. DjangoSettings._set_globals
            1. DjangoSettings.__init__
            2. <settings module>
        """
        stack = inspect.stack()
        settings_module_frame = stack[offset].frame
        settings_module_frame.f_locals.update(self.model_dump(by_alias=True))
